@@ -1,6 +1,6 @@
 from numpy import exp, ones_like, zeros_like, arange, multiply, \
      subtract, add, minimum, maximum, sign, c_, argmax, \
-     array, where
+     array, where, hstack
 from numpy import sum as npsum
 import matplotlib.pyplot as plt
 from math import isclose
@@ -680,9 +680,9 @@ class IT2FS(object):
         
         Returns a negated copy of the IT2FS.
         """
-        neg_it2fs = IT2FS(self.domain)
-        neg_it2fs.upper = subtract(1, self.upper)
-        neg_it2fs.lower = subtract(1, self.lower)
+        umf = lambda x, params: subtract(1, self.umf(x, params))
+        lmf = lambda x, params: subtract(1, self.lmf(x, params))
+        neg_it2fs = IT2FS(self.domain, umf, self.umf_params, lmf, self.lmf_params)
         return neg_it2fs
 
 
@@ -1150,9 +1150,11 @@ def meet(domain, it2fs1, it2fs2, t_norm):
     >>> it2fs3 = meet(domain, it2fs1, it2fs2, min_t_norm)
     >>> it2fs3.plot()
     """
-    it2fs = IT2FS(domain)
-    it2fs.upper = t_norm(it2fs1.upper, it2fs2.upper)
-    it2fs.lower = t_norm(it2fs1.lower, it2fs2.lower)
+    umf = lambda x, params: t_norm(it2fs1.umf(x, it2fs1.umf_params), 
+                                   it2fs2.umf(x, it2fs2.umf_params))
+    lmf = lambda x, params: t_norm(it2fs1.lmf(x, it2fs1.lmf_params), 
+                                   it2fs2.lmf(x, it2fs2.lmf_params))
+    it2fs = IT2FS(domain, umf, [], lmf, [])
     return it2fs
 
 
@@ -1198,9 +1200,11 @@ def join(domain, it2fs1, it2fs2, s_norm):
     >>> it2fs3 = join(domain, it2fs1, it2fs2, max_s_norm)
     >>> it2fs3.plot()
     """
-    it2fs = IT2FS(domain)
-    it2fs.upper = s_norm(it2fs1.upper, it2fs2.upper)
-    it2fs.lower = s_norm(it2fs1.lower, it2fs2.lower)
+    umf = lambda x, params: s_norm(it2fs1.umf(x, it2fs1.umf_params), 
+                                   it2fs2.umf(x, it2fs2.umf_params))
+    lmf = lambda x, params: s_norm(it2fs1.lmf(x, it2fs1.lmf_params), 
+                                   it2fs2.lmf(x, it2fs2.lmf_params))
+    it2fs = IT2FS(domain, umf, [], lmf, [])
     return it2fs
 
 def trim(intervals):
@@ -2076,9 +2080,9 @@ class IT2FLS(object):
     
     >>> domain = linspace(0., 1., 100)
     >>> 
-    >>> Small = IT2FS_Gaussian_UncertStd(domain, [0, 0.15, 0.1])
-    >>> Medium = IT2FS_Gaussian_UncertStd(domain, [0.5, 0.15, 0.1])
-    >>> Large = IT2FS_Gaussian_UncertStd(domain, [1., 0.15, 0.1])
+    >>> Small = IT2FS_Gaussian_UncertStd(domain, [0, 0.15, 0.1, 1.])
+    >>> Medium = IT2FS_Gaussian_UncertStd(domain, [0.5, 0.15, 0.1, 1.])
+    >>> Large = IT2FS_Gaussian_UncertStd(domain, [1., 0.15, 0.1, 1.])
     >>> IT2FS_plot(Small, Medium, Large, legends=["Small", "Medium", "large"])
     >>> 
     >>> myIT2FLS = IT2FLS()
@@ -2116,6 +2120,7 @@ class IT2FLS(object):
         self.rules = []
 
     def __repr__(self):
+        # TODO!
         pass
 
     def add_input_variable(self, name):
@@ -2306,14 +2311,16 @@ class IT2FLS(object):
                         u = t_norm(u, input_statement[1].umf(inputs[input_statement[0]], input_statement[1].umf_params))
                         l = t_norm(l, input_statement[1].lmf(inputs[input_statement[0]], input_statement[1].lmf_params))
                     for consequent in rule[1]:
-                        B_l = meet(domain, IT2FS(domain, const_mf, [u], const_mf, [l]), consequent[1], t_norm)
+                        B_l = meet(consequent[1].domain, 
+                                   IT2FS(consequent[1].domain, const_mf, [u], const_mf, [l]), 
+                                   consequent[1], t_norm)
                         B[consequent[0]].append(B_l)
-                C = {out: IT2FS(domain) for out in self.outputs}
+                C = {out: IT2FS(B[out][0].domain) for out in self.outputs}
                 TR = {}
                 for out in self.outputs:
                     for B_l in B[out]:
-                        C[out] = join(domain, C[out], B_l, s_norm)
-                    TR[out] = Centroid(C[out], alg_func, domain, alg_params=algorithm_params)
+                        C[out] = join(B_l.domain, C[out], B_l, s_norm)
+                    TR[out] = Centroid(C[out], alg_func, B_l.domain, alg_params=algorithm_params)
                 Cs.append(C)
                 TRs.append(TR)
             return Cs, TRs
@@ -2332,7 +2339,8 @@ class IT2FLS(object):
                         G[consequent[0]].append(consequent[1])
                 TR = {}
                 for out in self.outputs:
-                    TR[out] = CoSet(array(F), G[out], alg_func, domain, alg_params=algorithm_params)
+                    TR[out] = CoSet(array(F), G[out], alg_func, 
+                                    G[out][0].domain, alg_params=algorithm_params)
                 outputs.append(TR)
             return outputs
         elif method == "CoSum":
@@ -2345,11 +2353,13 @@ class IT2FLS(object):
                         u = t_norm(u, input_statement[1].umf(inputs[input_statement[0]], input_statement[1].umf_params))
                         l = t_norm(l, input_statement[1].lmf(inputs[input_statement[0]], input_statement[1].lmf_params))
                     for consequent in rule[1]:
-                        B_l = meet(domain, IT2FS(domain, const_mf, [u], const_mf, [l]), consequent[1], t_norm)
+                        B_l = meet(consequent[1].domain, 
+                                   IT2FS(consequent[1].domain, const_mf, [u], const_mf, [l]), 
+                                   consequent[1], t_norm)
                         B[consequent[0]].append(B_l)
                 TR = {}
                 for out in self.outputs:
-                    TR[out] = CoSum(B[out], alg_func, domain, alg_params=algorithm_params)
+                    TR[out] = CoSum(B[out], alg_func, B[out][0].domain, alg_params=algorithm_params)
                 outputs.append(TR)
             return outputs
         elif method == "Height":
@@ -2362,11 +2372,14 @@ class IT2FLS(object):
                         u = t_norm(u, input_statement[1].umf(inputs[input_statement[0]], input_statement[1].umf_params))
                         l = t_norm(l, input_statement[1].lmf(inputs[input_statement[0]], input_statement[1].lmf_params))
                     for consequent in rule[1]:
-                        B_l = meet(domain, IT2FS(domain, const_mf, [u], const_mf, [l]), consequent[1], t_norm)
+                        B_l = meet(consequent[1].domain, 
+                                   IT2FS(consequent[1].domain, const_mf, [u], const_mf, [l]), 
+                                   consequent[1], t_norm)
                         B[consequent[0]].append(B_l)
                 TR = {}
                 for out in self.outputs:
-                    TR[out] = Height(B[out], alg_func, domain, alg_params=algorithm_params)
+                    TR[out] = Height(B[out], alg_func, 
+                                     B[out][0].domain, alg_params=algorithm_params)
                 outputs.append(TR)
             return outputs
         elif method == "ModiHe":
@@ -2379,11 +2392,14 @@ class IT2FLS(object):
                         u = t_norm(u, input_statement[1].umf(inputs[input_statement[0]], input_statement[1].umf_params))
                         l = t_norm(l, input_statement[1].lmf(inputs[input_statement[0]], input_statement[1].lmf_params))
                     for consequent in rule[1]:
-                        B_l = meet(domain, IT2FS(domain, const_mf, [u], const_mf, [l]), consequent[1], t_norm)
+                        B_l = meet(consequent[1].domain, 
+                                   IT2FS(consequent[1].domain, const_mf, [u], const_mf, [l]), 
+                                   consequent[1], t_norm)
                         B[consequent[0]].append(B_l)
                 TR = {}
                 for out in self.outputs:
-                    TR[out] = ModiHe(B[out], method_params, alg_func, domain, alg_params=algorithm_params)
+                    TR[out] = ModiHe(B[out], method_params, alg_func, 
+                                     B[out][0].domain, alg_params=algorithm_params)
                 outputs.append(TR)
             return outputs
         else:
@@ -2496,14 +2512,16 @@ class IT2FLS(object):
                     u = t_norm(u, input_statement[1].umf(inputs[input_statement[0]], input_statement[1].umf_params))
                     l = t_norm(l, input_statement[1].lmf(inputs[input_statement[0]], input_statement[1].lmf_params))
                 for consequent in rule[1]:
-                    B_l = meet(domain, IT2FS(domain, const_mf, [u], const_mf, [l]), consequent[1], t_norm)
+                    B_l = meet(consequent[1].domain, 
+                               IT2FS(consequent[1].domain, const_mf, [u], const_mf, [l]), 
+                               consequent[1], t_norm)
                     B[consequent[0]].append(B_l)
-            C = {out: IT2FS(domain) for out in self.outputs}
+            C = {out: IT2FS(B[out][0].domain) for out in self.outputs}
             TR = {}
             for out in self.outputs:
                 for B_l in B[out]:
-                    C[out] = join(domain, C[out], B_l, s_norm)
-                TR[out] = Centroid(C[out], alg_func, domain, alg_params=algorithm_params)
+                    C[out] = join(B_l.domain, C[out], B_l, s_norm)
+                TR[out] = Centroid(C[out], alg_func, B_l.domain, alg_params=algorithm_params)
             return C, TR
         elif method == "CoSet":
             F = []
@@ -2519,7 +2537,8 @@ class IT2FLS(object):
                     G[consequent[0]].append(consequent[1])
             TR = {}
             for out in self.outputs:
-                TR[out] = CoSet(array(F), G[out], alg_func, domain, alg_params=algorithm_params)
+                TR[out] = CoSet(array(F), G[out], alg_func, 
+                                G[out][0].domain, alg_params=algorithm_params)
             return TR
         elif method == "CoSum":
             B = {out: [] for out in self.outputs}
@@ -2530,11 +2549,14 @@ class IT2FLS(object):
                     u = t_norm(u, input_statement[1].umf(inputs[input_statement[0]], input_statement[1].umf_params))
                     l = t_norm(l, input_statement[1].lmf(inputs[input_statement[0]], input_statement[1].lmf_params))
                 for consequent in rule[1]:
-                    B_l = meet(domain, IT2FS(domain, const_mf, [u], const_mf, [l]), consequent[1], t_norm)
+                    B_l = meet(consequent[1].domain, 
+                               IT2FS(consequent[1].domain, const_mf, [u], const_mf, [l]), 
+                               consequent[1], t_norm)
                     B[consequent[0]].append(B_l)
             TR = {}
             for out in self.outputs:
-                TR[out] = CoSum(B[out], alg_func, domain, alg_params=algorithm_params)
+                TR[out] = CoSum(B[out], alg_func, 
+                                B[out][0].domain, alg_params=algorithm_params)
             return TR
         elif method == "Height":
             B = {out: [] for out in self.outputs}
@@ -2545,11 +2567,14 @@ class IT2FLS(object):
                     u = t_norm(u, input_statement[1].umf(inputs[input_statement[0]], input_statement[1].umf_params))
                     l = t_norm(l, input_statement[1].lmf(inputs[input_statement[0]], input_statement[1].lmf_params))
                 for consequent in rule[1]:
-                    B_l = meet(domain, IT2FS(domain, const_mf, [u], const_mf, [l]), consequent[1], t_norm)
+                    B_l = meet(consequent[1].domain, 
+                               IT2FS(consequent[1].domain, const_mf, [u], const_mf, [l]), 
+                               consequent[1], t_norm)
                     B[consequent[0]].append(B_l)
             TR = {}
             for out in self.outputs:
-                TR[out] = Height(B[out], alg_func, domain, alg_params=algorithm_params)
+                TR[out] = Height(B[out], alg_func, 
+                                 B[out][0].domain, alg_params=algorithm_params)
             return TR
         elif method == "ModiHe":
             B = {out: [] for out in self.outputs}
@@ -2560,36 +2585,465 @@ class IT2FLS(object):
                     u = t_norm(u, input_statement[1].umf(inputs[input_statement[0]], input_statement[1].umf_params))
                     l = t_norm(l, input_statement[1].lmf(inputs[input_statement[0]], input_statement[1].lmf_params))
                 for consequent in rule[1]:
-                    B_l = meet(domain, IT2FS(domain, const_mf, [u], const_mf, [l]), consequent[1], t_norm)
+                    B_l = meet(consequent[1].domain, 
+                               IT2FS(consequent[1].domain, const_mf, [u], const_mf, [l]), 
+                               consequent[1], t_norm)
                     B[consequent[0]].append(B_l)
             TR = {}
             for out in self.outputs:
-                TR[out] = ModiHe(B[out], method_params, alg_func, domain, alg_params=algorithm_params)
+                TR[out] = ModiHe(B[out], method_params, alg_func, 
+                                 B[out][0].domain, alg_params=algorithm_params)
             return TR
         else:
             raise ValueError("The method " + method + " is not implemented yet!")
 
 
+class TSK:
 
+    def __init__(self, t_norm, s_norm):
+        self.inputs = []
+        self.outputs = []
+        self.rules = []
+        self.__t_norm = t_norm
+        self.__s_norm = s_norm
+    
+    def __repr__(self):
+        # TODO!
+        pass
 
+    def add_input_variable(self, name):
+        """
+        Adds new input variable name.
+        
+        Parameters
+        ----------
+        
+        name:
+            str
+            
+            Name of the new input variable as a str.
+        """
+        self.inputs.append(name)
 
+    def add_output_variable(self, name):
+        """
+        Adds new output variable name.
+        
+        Parameters
+        ----------
+        
+        name:
+            str
+            
+            Name of the new output variable as a str.
+        """
+        self.outputs.append(name)
 
+    def add_rule(self, antecedent, consequent):
+        """
+        Adds new rule to the rule base of the IT2FLS.
+        
+        Parameters
+        ----------
+        
+        antecedent:
+            List of tuples
+            
+            Antecedent is a list of tuples in which each tuple indicates 
+            assignement of a variable to an IT2FS. First element of the 
+            tuple must be input variable name as str, and the second 
+            element of the tuple must be an IT2FS.
+            
+        consequent:
+            List of tuples
+            
+            Consequent is a list of tuples in which each tuple indicates 
+            assignement of a variable to an output state. First element of the 
+            tuple must be output vriable name as str, and the second element 
+            of the tuple must be a dictionary. This dictionary shows the 
+            output polynomial in the case of the rule. For example let an 
+            output polynomial be as 2 x1 + 4 x2 + 5. Then the dictionary for 
+            this case would be {"const":5., "x1":2., "x2":4.}. Note that this 
+            is written for an IT2 TSK FLS with two inputs, named x1 and x2.
+        
+        Example
+        -------
+        
+        Assume that we are going to simulate an IT2 TSK FLS with two inputs 
+        named x1 and x2. Our rule base is defined as below:
+            
+            * IF x1 is Small and x2 is Small THEN y1 = x1 + x2 + 1 and y2 = 2 x1 - x2 + 1
+            * IF x1 is Small and x2 is Big THEN y1 = 1.5 x1 + 0.5 x2 + 0.5 and y2 = 1.5 x1 - 0.5 x2 + 0.5
+            * IF x1 is Big and x2 is Small THEN y1 = 2. x1 + 0.1 x2 - 0.2 and y2 = 0.5 x1 + 0.1 x2
+            * IF x1 is Big and x2 is Big THEN y1 = 4. x1 -0.5 x2 - 1 and y2 = -0.5 x1 + x2 - 0.5
+        
+        Then these rules can be added to the system using the codes below:
+        
+        >>> myIT2FLS.add_rule([("x1", Small), ("x2", Small)], 
+                              [("y1", {"const":1., "x1":1., "x2":1.}), 
+                               ("y2", {"const":1., "x1":2., "x2":-1.})])
+        >>> myIT2FLS.add_rule([("x1", Small), ("x2", Big)], 
+                              [("y1", {"const":0.5, "x1":1.5, "x2":0.5}), 
+                               ("y2", {"const":0.5, "x1":1.5, "x2":-0.5})])
+        >>> myIT2FLS.add_rule([("x1", Big), ("x2", Small)], 
+                              [("y1", {"const":-0.2, "x1":2., "x2":0.1}), 
+                               ("y2", {"const":0., "x1":0.5, "x2":0.1})])
+        >>> myIT2FLS.add_rule([("x1", Big), ("x2", Big)], 
+                              [("y1", {"const":-1., "x1":4., "x2":-0.5}), 
+                               ("y2", {"const":-0.5, "x1":-0.5, "x2":1.})])
+        
+        """
+        self.rules.append((antecedent, consequent))
+    
+    def copy(self):
+        """
+        Returns a copy of the IT2FLS.
+        """
+        o = TSK(self.__t_norm, self.__s_norm)
+        o.inputs = self.inputs.copy()
+        o.outputs = self.outputs.copy()
+        o.rules = self.rules.copy()
+        return o
 
+    def evaluate(self, inputs):
+        """
+        Evaluates the IT2 TSK FLS based on the crisp inputs given by the user.
 
+        Parameters
+        ----------
+        inputs: 
+            dictionary
+            
+            Inputs is a dictionary, which the keys are input variable 
+            names as str and the values are the crisp value of the inputs to 
+            be evaluated.
 
+        Returns
+        -------
+        O: 
+            dictionary
+            
+            The output is a dictionary, which the keys are output variable 
+            names as str and the values are the crisp output of the system.
 
+        """
+        F = []
+        B = {out: [] for out in self.outputs}
+        for rule in self.rules:
+            u = 1
+            l = 1
+            for input_statement in rule[0]:
+                u = self.__t_norm(u, input_statement[1].umf(inputs[input_statement[0]], input_statement[1].umf_params))
+                l = self.__t_norm(l, input_statement[1].lmf(inputs[input_statement[0]], input_statement[1].lmf_params))
+            F.append([l, u])
+            for consequent in rule[1]:
+                B[consequent[0]].append(consequent[1]["const"])
+                for input in self.inputs:
+                    # self.inputs -> Name of the input variables ...
+                    # inputs -> Dict of input and value pairs ...
+                    # consequent[1] -> dict of input and coefficient pairs ...
+                    B[consequent[0]][-1] += consequent[1][input] * inputs[input]
+        O = {}
+        for output in self.outputs:
+            y = array(B[output]).reshape((len(B[output]), 1))
+            intervals = hstack([y, y, array(F)])
+            o = EIASC_algorithm(intervals)
+            O[output] = crisp(o)
+        return O
 
+class Mamdani:
+    """
+    Interval Type 2 Mamadani Fuzzy Logic System.
+    
+    Parameters
+    ----------
+    Parameters of the constructor function:
+        
+    t_norm:
+        function
+        
+        T-norm operator which would be used in FLS.
+    
+    s_norm:
+        function
+        
+        S-norm operator which would be used in FLS.
+    
+    method="Centroid":
+        str
+        
+        Indicates the type reduction method name and should be one 
+        of the methods listed below:
+            Centroid, CoSet, CoSum, Height, and ModiHe.
+    
+    method_params=None:
+        List
+        
+        Parameters of the type reduction method, if needed.
+    
+    algorithm="EIASC":
+        str
+        
+        Indicates the type reduction algorithm name and should be 
+        one of the algorithms listed below:
+            KM, EKM, WEKM, TWEKM, EIASC, WM, BMM, LBMM, and NT.
+    
+    algorithm_params=None:
+        List
+        
+        Parameters of the type reduction algorithm, if needed.
+    
+    Members
+    -------
+    
+    inputs:
+        List of str
+        
+        List of names of inputs as str
+    
+    output:
+        List of str
+        
+        List of names ot outputs as str
+        
+    rules:
+        List of tuples (antecedent, consequent)
+        
+        List of rules which each rule is defined as a 
+        tuple (antecedent, consequent)
+        
+        Both antacedent and consequent are lists of tuples. Each tuple 
+        of this list shows assignement of a variable to an IT2FS. 
+        First element of the tuple must be variable name (input or output) 
+        as a str and the second element must be an IT2FS. 
+    
+    Functions
+    ---------
+    
+    add_input_variable:
+        
+        Adds an input variable to the inputs list of the IT2FLS.
+    
+    add_output_variable:
+        
+        Adds an output variable to the outputs list of the IT2FLS.
+    
+    add_rule:
+        
+        Adds a rule to the rules list of the IT2FLS.
+    
+    copy:
+        
+        Returns a copy of the IT2FLS.
+    
+    evaluate:
+        
+        Evaluates the IT2FLS's output for a specified crisp input. This function 
+        is selected based on the constructor function parameter, method, among 5 
+        private functions below:
+        __Mamdani_Centroid, __Mamdani_CoSet, __Mamdani_CoSum, __Mamdani_Height, and 
+        __Mamdani_ModiHe.
+        The only input of the evaluate function is a dictionary named inputs 
+        in which the keys are input variable names as str and the values are 
+        the crisp value of inputs to be evaluated.
+        The output of the evaluate function is depended on the method selected 
+        while constructing the class. For more information, please refer to 
+        the examples.
+    
+    """
+    def __init__(self, t_norm, s_norm, 
+                 method="Centroid", method_params=None, 
+                 algorithm=EIASC_algorithm, algorithm_params=None):
+        self.inputs = []
+        self.outputs = []
+        self.rules = []
+        self.__t_norm = t_norm
+        self.__s_norm = s_norm
+        self.__method = method
+        self.__method_params = method_params
+        self.__algorithm = algorithm
+        self.__algorithm_params = algorithm_params
+        if method == "Centroid":
+            self.evaluate = self.__Mamdani_Centroid
+        elif method == "CoSet":
+            self.evaluate = self.__Mamdani_CoSet
+        elif method == "CoSum":
+            self.evaluate = self.__Mamdani_CoSum
+        elif method == "Height":
+            self.evaluate = self.__Mamdani_Height
+        elif method == "ModiHe":
+            self.evaluate = self.__Mamdani_ModiHe
+        else:
+            raise ValueError("The method " + method + " is not implemented yet!")
 
+    def __repr__(self):
+        # TODO!
+        pass
 
+    def add_input_variable(self, name):
+        """
+        Adds new input variable name.
+        
+        Parameters
+        ----------
+        
+        name:
+            str
+            
+            Name of the new input variable as a str.
+        """
+        self.inputs.append(name)
 
+    def add_output_variable(self, name):
+        """
+        Adds new output variable name.
+        
+        Parameters
+        ----------
+        
+        name:
+            str
+            
+            Name of the new output variable as a str.
+        """
+        self.outputs.append(name)
 
-
-
-
-
-
-
-
+    def add_rule(self, antecedent, consequent):
+        """
+        Adds new rule to the rule base of the IT2FLS.
+        
+        Parameters
+        ----------
+        
+        antecedent:
+            List of tuples
+            
+            Antecedent is a list of tuples in which each tuple indicates 
+            assignement of a variable to an IT2FS. First element of the 
+            tuple must be input variable name as str, and the second 
+            element of the tuple must be an IT2FS.
+            
+        consequent:
+            List of tuples
+            
+            Consequent is a list of tuples in which each tuple indicates 
+            assignement of a variable to an IT2FS. First element of the 
+            tuple must be output variable name as str, and the second 
+            element of the tuple must be an IT2FS.
+            
+        """
+        self.rules.append((antecedent, consequent))
+    
+    def copy(self):
+        """
+        Returns a copy of the IT2FLS.
+        """
+        o = Mamdani(self.__t_norm, self.__s_norm, 
+                    self.__method, self.__method_params, 
+                    self.__algorithm, self.__algorithm_params)
+        o.inputs = self.inputs.copy()
+        o.outputs = self.outputs.copy()
+        o.rules = self.rules.copy()
+        return o
+    
+    def __Mamdani_Centroid(self, inputs):
+        B = {out: [] for out in self.outputs}
+        for rule in self.rules:
+            u = 1
+            l = 1
+            for input_statement in rule[0]:
+                u = self.__t_norm(u, input_statement[1].umf(inputs[input_statement[0]], input_statement[1].umf_params))
+                l = self.__t_norm(l, input_statement[1].lmf(inputs[input_statement[0]], input_statement[1].lmf_params))
+            for consequent in rule[1]:
+                B_l = meet(consequent[1].domain, 
+                           IT2FS(consequent[1].domain, const_mf, [u], const_mf, [l]), 
+                           consequent[1], self.__t_norm)
+                B[consequent[0]].append(B_l)
+        C = {out: IT2FS(B[out][0].domain) for out in self.outputs}
+        TR = {}
+        for out in self.outputs:
+            for B_l in B[out]:
+                C[out] = join(B_l.domain, C[out], B_l, self.__s_norm)
+            TR[out] = Centroid(C[out], self.__algorithm, B_l.domain, 
+                               alg_params=self.__algorithm_params)
+        return C, TR
+    
+    def __Mamdani_CoSet(self, inputs):
+        F = []
+        G = {out: [] for out in self.outputs}
+        for rule in self.rules:
+            u = 1
+            l = 1
+            for input_statement in rule[0]:
+                u = self.__t_norm(u, input_statement[1].umf(inputs[input_statement[0]], input_statement[1].umf_params))
+                l = self.__t_norm(l, input_statement[1].lmf(inputs[input_statement[0]], input_statement[1].lmf_params))
+            F.append([l, u])
+            for consequent in rule[1]:
+                G[consequent[0]].append(consequent[1])
+        TR = {}
+        for out in self.outputs:
+            TR[out] = CoSet(array(F), G[out], self.__algorithm, 
+                            G[out][0].domain, alg_params=self.__algorithm_params)
+        return TR
+    
+    def __Mamdani_CoSum(self, inputs):
+        B = {out: [] for out in self.outputs}
+        for rule in self.rules:
+            u = 1
+            l = 1
+            for input_statement in rule[0]:
+                u = self.__t_norm(u, input_statement[1].umf(inputs[input_statement[0]], input_statement[1].umf_params))
+                l = self.__t_norm(l, input_statement[1].lmf(inputs[input_statement[0]], input_statement[1].lmf_params))
+            for consequent in rule[1]:
+                B_l = meet(consequent[1].domain, 
+                           IT2FS(consequent[1].domain, const_mf, [u], const_mf, [l]), 
+                           consequent[1], self.__t_norm)
+                B[consequent[0]].append(B_l)
+        TR = {}
+        for out in self.outputs:
+            TR[out] = CoSum(B[out], self.__algorithm, 
+                            B[out][0].domain, alg_params=self.__algorithm_params)
+        return TR
+    
+    def __Mamdani_Height(self, inputs):
+        B = {out: [] for out in self.outputs}
+        for rule in self.rules:
+            u = 1
+            l = 1
+            for input_statement in rule[0]:
+                u = self.__t_norm(u, input_statement[1].umf(inputs[input_statement[0]], input_statement[1].umf_params))
+                l = self.__t_norm(l, input_statement[1].lmf(inputs[input_statement[0]], input_statement[1].lmf_params))
+            for consequent in rule[1]:
+                B_l = meet(consequent[1].domain, 
+                           IT2FS(consequent[1].domain, const_mf, [u], const_mf, [l]), 
+                           consequent[1], self.__t_norm)
+                B[consequent[0]].append(B_l)
+        TR = {}
+        for out in self.outputs:
+            TR[out] = Height(B[out], self.__algorithm, 
+                             B[out][0].domain, alg_params=self.__algorithm_params)
+        return TR
+    
+    def __Mamdani_ModiHe(self, inputs):
+        B = {out: [] for out in self.outputs}
+        for rule in self.rules:
+            u = 1
+            l = 1
+            for input_statement in rule[0]:
+                u = self.__t_norm(u, input_statement[1].umf(inputs[input_statement[0]], input_statement[1].umf_params))
+                l = self.__t_norm(l, input_statement[1].lmf(inputs[input_statement[0]], input_statement[1].lmf_params))
+            for consequent in rule[1]:
+                B_l = meet(consequent[1].domain, 
+                           IT2FS(consequent[1].domain, const_mf, [u], const_mf, [l]), 
+                           consequent[1], self.__t_norm)
+                B[consequent[0]].append(B_l)
+        TR = {}
+        for out in self.outputs:
+            TR[out] = ModiHe(B[out], self.__method_params, self.__algorithm, 
+                             B[out][0].domain, alg_params=self.__algorithm_params)
+        return TR
+    
+    
 
 
 
